{
This is a component for placing icons in the notification area
of the taskbar (aka. the traybar).

This component is an improvement of the TTraybar component made by
JH (?) (j.h.dewaalmalefijt@wbmt.tudelft.nl). Praise and thanks go
to him. I corrected some minor bugs and added some new properties
and methods.

I also nicked a bit of the logic in the tray icon's event handling
(procedure HandleMessage) from another TTrayIcon component by
Tempest software (don't know their address). I don't believe I've
violated any copyrights, but in any case this component is
freeware. Feel free to use and improve the component. I would
like to hear what you think.

Troels Jakobsen - tjak@get2net.dk
}

unit TrayIcon;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Menus, ShellApi;

const
  { User-defined message sent from the icon. Some low user-defined
    messages are used by Windows itself! (WM_USER+1 = DM_SETDEFID) }
  WM_TRAYNOTIFY = WM_USER + 1024;
  IconID = 1;

type
  TTrayIcon = class(TComponent)
  private
    tnid: TNotifyIconData;
//    FEnabled: Boolean;
    FIcon: TIcon;
    FIconVisible: Boolean;
    FHint: String;
    FShowHint: Boolean;
    FPopupMenu: TPopupMenu;
    FLeftPopup: Boolean;
    FOnClick,
    FOnDblClick: TNotifyEvent;
    FOnMouseDown,
    FOnMouseUp: TMouseEvent;
    FOnMouseMove: TMouseMoveEvent;
    FStartMinimized: Boolean;
    ShowFirst: Boolean;
    FClicked: Boolean;
    procedure HandleMessage(var Msg: TMessage);
    function InitIcon: Boolean;
    procedure SetIcon(Value: TIcon);
    procedure SetIconVisible(Value: Boolean);
    procedure SetHint(Value: String);
    procedure SetShowHint(Value: Boolean);
    procedure PopupAtCursor;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ShowMainForm;
    procedure HideMainForm;
  protected
    procedure Loaded; override;
    function ShowIcon: Boolean; virtual;
    function HideIcon: Boolean; virtual;
    function ModifyIcon: Boolean; virtual;
    procedure Click; dynamic;
    procedure DblClick; dynamic;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); dynamic;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); dynamic;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); dynamic;
  published
    // Properties:
    property Hint: String read FHint write SetHint;
    property ShowHint: Boolean read FShowHint write SetShowHint;
    property Icon: TIcon read FIcon write SetIcon stored True;
    property IconVisible: Boolean read FIconVisible write SetIconVisible
      default True;
    property PopupMenu: TPopupMenu read FPopupMenu write FPopupMenu;
    property LeftPopup: Boolean read FLeftPopup write FLeftPopup
      default False;
    property StartMinimized: Boolean read FStartMinimized write FStartMinimized
      default False;         // Main form minimized on appl. start-up?
    // Events:
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
  end;

procedure Register;

implementation

{--------------------- TTrayIcon ----------------------}

constructor TTrayIcon.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIconVisible := True;      // True is the default property value
  ShowFirst := True;         // The main form has not been shown before

  // Create tray icon itself using never changing fields of tnid
  FIcon := TIcon.Create;
  tnid.cbSize := SizeOf(TNotifyIconData);
  // tnid.wnd points to procedure to receive callback messages from the icon
  tnid.Wnd := AllocateHWnd(HandleMessage);
  // Just a number for the new icon
  tnid.uId := IconID;
  // Icon, message and tiptext members are/must be valid
  tnid.uFlags := NIF_ICON + NIF_MESSAGE + NIF_TIP;
  // Message to send to tnid.wnd when mouse event occurs
  tnid.uCallbackMessage := WM_TRAYNOTIFY;
//  tnid.hIcon := FIcon.Handle;
end;


destructor TTrayIcon.Destroy;
begin
  SetIconVisible(False);     // Remove the icon from the tray
  FIcon.Free;                // Free the icon
  DeallocateHWnd(tnid.Wnd);  // Free the handle of tray window
  inherited Destroy;
end;


procedure TTrayIcon.Loaded;
{ This method is called when all properties of the component have been
  initialized. The method SetIconVisible must be called here, after the
  tray icon (FIcon) has loaded itself. Otherwise, the tray icon will
  be blank (no icon image). }
begin
  inherited Loaded;	     // Always call inherited Loaded first
  SetIconVisible(FIconVisible);

  if StartMinimized then
  begin
    Application.ShowMainForm := False;
    ShowWindow(Application.Handle, SW_HIDE);
  end;
end;


{ Message handler for the shell notification window (tray icon).
  Most messages are passed through WM_TRAYNOTIFY. Use lParam to
  get the actual message, eg. WM_MOUSEMOVE.

  Send the usual Delphi events for the mouse messages. Also
  interpolate the OnClick event when the user clicks the left
  button, and popup the menu, if there is one, for left and right
  mouse down events. }

procedure TTrayIcon.HandleMessage(var Msg: TMessage);
// Handle incoming messages from the traybar icon

  function ShiftState: TShiftState;
  // Return the state of the shift, ctrl, and alt keys
  begin
    Result := [];
    if GetKeyState(VK_SHIFT) < 0 then
      Include(Result, ssShift);
    if GetKeyState(VK_CONTROL) < 0 then
      Include(Result, ssCtrl);
    if GetKeyState(VK_MENU) < 0 then
      Include(Result, ssAlt);
  end;

var
  Pt: TPoint;
  Shift: TShiftState;
begin
  if Msg.Msg = WM_TRAYNOTIFY then
  // Take action if a message from the icon comes through
  begin
    case Msg.lParam of
    WM_MOUSEMOVE:
      begin
        Shift := ShiftState;
        GetCursorPos(Pt);
        MouseMove(Shift, Pt.X, Pt.Y);
      end;
    WM_LBUTTONDOWN:
      begin
        Shift := ShiftState + [ssLeft];
        GetCursorPos(Pt);
        MouseDown(mbLeft, Shift, Pt.X, Pt.Y);
        FClicked := True;
        if FLeftPopup then
          PopupAtCursor;
      end;
    WM_RBUTTONDOWN:
      begin
        Shift := ShiftState + [ssRight];
        GetCursorPos(Pt);
        MouseDown(mbRight, Shift, Pt.X, Pt.Y);
        PopupAtCursor;
      end;
    WM_MBUTTONDOWN:
      begin
        Shift := ShiftState + [ssMiddle];
        GetCursorPos(Pt);
        MouseDown(mbMiddle, Shift, Pt.X, Pt.Y);
      end;
    WM_LBUTTONUP:
      begin
        Shift := ShiftState + [ssLeft];
        GetCursorPos(Pt);
        if FClicked then     // Then WM_LBUTTONDOWN was called before
        begin
          FClicked := False;
          Click;
        end;
        MouseUp(mbLeft, Shift, Pt.X, Pt.Y);
      end;
    WM_RBUTTONUP:
      begin
        Shift := ShiftState + [ssRight];
        GetCursorPos(Pt);
        MouseUp(mbRight, Shift, Pt.X, Pt.Y);
      end;
    WM_MBUTTONUP:
      begin
        Shift := ShiftState + [ssMiddle];
        GetCursorPos(Pt);
        MouseUp(mbMiddle, Shift, Pt.X, Pt.Y);
      end;
    WM_LBUTTONDBLCLK, WM_RBUTTONDBLCLK, WM_MBUTTONDBLCLK: DblClick;
    end;
  end
  else        // Messages that didn't go through the icon
    case Msg.Msg of
      WM_QUERYENDSESSION: Msg.Result := 1;
      { Evaluate WM_QUERYENDSESSION message to tell Windows that the
        icon will stop executing if user requests a shutdown (Msg.Result
        must not return 0, or the system will not be able to shut down). }
    else      // Handle all other messages with the default handler
      DefWindowProc(tnid.Wnd, Msg.Msg, Msg.wParam, Msg.lParam);
//      Msg.Result := DefWindowProc(tnid.Wnd, Msg.Msg, Msg.wParam, Msg.lParam);
    end;
end;


function TTrayIcon.InitIcon: Boolean;
// Fill tnid with changing data, like icon and tooltip
begin
  if not (csDesigning in ComponentState){Not in Designmode} and
     not (csLoading in ComponentState) {And not loading component} then
    // Above checks aren't really necessary
    begin
      tnid.hIcon := FIcon.Handle;
      if (FHint <> '') and (FShowHint) then
        StrLCopy(tnid.szTip, PChar(FHint), SizeOf(tnid.szTip))
        // StrLCopy must be used since szTip is only 64 bytes (dumb)
      else
        tnid.szTip := '';
      Result := True;
    end
  else
    Result := False;
end;


procedure TTrayIcon.SetIcon(Value: TIcon);
begin
  FIcon.Assign(Value);
end;


procedure TTrayIcon.SetIconVisible(Value: Boolean);
begin
  if Value then
    ShowIcon
  else
    HideIcon;
  FIconVisible := Value;     // Not really necessary
end;


procedure TTrayIcon.SetHint(Value: String);
begin
  FHint := Value;
  ModifyIcon;
end;


procedure TTrayIcon.SetShowHint(Value: Boolean);
begin
  FShowHint := Value;
  ModifyIcon;
end;


function TTrayIcon.ShowIcon: Boolean;
// Add/show the icon on the tray
begin
  if InitIcon then
  begin
    Result := Shell_NotifyIcon(NIM_ADD, @tnid);
    FIconVisible := True;
  end;
end;


function TTrayIcon.HideIcon: Boolean;
// Remove/hide the icon from the tray
begin
  if InitIcon then
  begin
    Result := Shell_NotifyIcon(NIM_DELETE, @tnid);
    FIconVisible := False;
  end;
end;


function TTrayIcon.ModifyIcon: Boolean;
// Change icon or tooltip if icon already placed
begin
  if InitIcon then
    Result := Shell_NotifyIcon(NIM_MODIFY, @tnid);
end;


procedure TTrayIcon.ShowMainForm;
var
  I: Integer;
begin
  Application.ShowMainForm := True;
  // Show application's TASKBAR icon (not the traybar icon)
  ShowWindow(Application.Handle, SW_RESTORE);
  // Show the form itself
  ShowWindow(Application.MainForm.Handle, SW_RESTORE);

  { If the main form has not been shown before (if StartMinimized
    was true (Application.ShowMainForm was false on startup)),
    it's necessary to force the form's controls to show, as they
    have been created invisible (regardless of the value of their
    Visible property). This is done via ShowWindow. }
  { By the way: TForm.Position has no effect if StartMinimized
    is true. Kind of stupid. }
  if ShowFirst then          // This block is only executed once
  begin
    for I := 0 to Application.MainForm.ComponentCount -1 do
      if Application.MainForm.Components[I] is TWinControl then
        with Application.MainForm.Components[I] as TWinControl do
          if Visible then
            ShowWindow(Handle, SW_SHOWDEFAULT);
    ShowFirst := False;      // The main form has now been shown
  end;
end;


procedure TTrayIcon.HideMainForm;
begin
  Application.ShowMainForm := False;
  // Hide application's TASKBAR icon (not the traybar icon)
  ShowWindow(Application.Handle, SW_HIDE);
  // Hide the form itself
  ShowWindow(Application.MainForm.Handle, SW_HIDE);
end;


procedure TTrayIcon.PopupAtCursor;
var
  CursorPos: TPoint;
begin
  if Assigned(PopupMenu) then
    if PopupMenu.AutoPopup then
      if GetCursorPos(CursorPos) then
      begin
        PopupMenu.PopupComponent := Self;
        PopupMenu.Popup(CursorPos.X, CursorPos.Y);
      end;
end;


procedure TTrayIcon.Click;
begin
  // Call the method that user has assigned to the event (if any)
  if Assigned(FOnClick) then
    FOnClick(Self);
end;


procedure TTrayIcon.DblClick;
begin
  if Assigned(FOnDblClick) then
    FOnDblClick(Self);
end;


procedure TTrayIcon.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if Assigned(FOnMouseDown) then
    FOnMouseDown(Self, Button, Shift, X, Y);
end;


procedure TTrayIcon.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if Assigned(FOnMouseUp) then
    FOnMouseUp(Self, Button, Shift, X, Y);
end;


procedure TTrayIcon.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnMouseMove) then
    FOnMouseMove(Self, Shift, X, Y);
end;


procedure Register;
begin
  RegisterComponents('Custom', [TTrayIcon]);
end;

end.

